"""interpret event traces as test activity

this file serves as an example for the xunitgen module

event traces are similar/compatible with those produced by uu.spdr

https://github.com/uucidl/uu.spdr
"""
from argparse import ArgumentParser

import os
import re

from socket import gethostname

from xunitgen import XunitDestination, EventReceiver, toxml


def parse_trace(line):
    main_re = r'^TT01 ([0-9]+) ([0-9]+) ([0-9]+) ("(?:[^\\"]|\\.)*") ("(?:[^\\"]|\\.)*") ("[A-Z]")(.*)$'
    match = re.match(main_re, line)
    if match:
        def get(i):
            return match.group(i)

        def unquote_str(string):
            return eval(string)

        trace = dict(
            ts=int(get(1)),
            pid=int(get(2)),
            tid=int(get(3)),
            cat=unquote_str(get(4)),
            name=unquote_str(get(5)),
            ph=unquote_str(get(6)),
            args={},
        )

        arg_str = get(7)
        if arg_str:
            for match in re.finditer(r' ("(?:[^\\"]|\\.)*") ((?:"(?:[^\\"]|\\.)*")|(?:[0-9]+\.[0-9]*)|(?:[0-9]+))', arg_str):
                name = unquote_str(get(1))
                value_str = get(2)
                if re.match(r'^"(?:[^\\"]|\\.)*"$', value_str):
                    value = unquote_str(value_str)
                elif re.match(r'^[0-9]+\.[0-9]*$', value_str):
                    value = float(value_str)
                elif re.match(r'^[0-9]+$', value_str):
                    value = int(value_str)

                trace['args'][name] = value

            if not trace['args']:
                raise Exception('Could not parse args %r' % arg_str)

        return trace

    raise Exception('Could not parse %r' % line)


def gather_test_results(traces):
    MICROS_TO_S = 1000000.0
    receiver = EventReceiver()

    for trace in traces:
        ts_seconds = trace['ts'] / MICROS_TO_S
        if trace['cat'] == 'test':
            if trace['ph'] == 'E' and receiver.current_case.name == trace['name']:
                receiver.end_case(trace['name'], ts_seconds)
            elif trace['ph'] == 'B' and receiver.current_case is None:
                receiver.begin_case(
                    trace['name'], ts_seconds, os.path.splitext(
                        trace['args']['filename'])[0].replace(os.sep, '.')
                )
            elif trace['ph'] == 'I' and trace['name'] == 'failure':
                receiver.failure(trace['args'][
                                 'reason'], trace['args']['lineno'])

    return receiver.results()


def main():
    parser = ArgumentParser()
    parser.add_argument("dst_xunit_file")
    parser.add_argument("src_trace_log")

    args = parser.parse_args()

    destination = XunitDestination(os.path.dirname(os.path.abspath(args.dst_xunit_file)))
    xml_filepath = os.path.splitext(os.path.basename(args.dst_xunit_file))[0]
    with open(args.src_trace_log) as file:
        def parse_line(line_i, line):
            try:
                return parse_trace(line)
            except Exception as e:
                raise Exception('%s:%d: error: %r' % (
                    args.src_trace_log, i, e))

        test_results = gather_test_results(
            [parse_line(i, line) for i, line in enumerate(file.readlines())]
        )

    destination.write_reports(xml_filepath, 'testsuite', test_results)


if __name__ == "__main__":
    main()



import sys
import time

from contextlib import contextmanager

from .main import EventReceiver

class Recorder(object):
    """Use this class to record the result of running python code as a xunit xml

    It allows you to record a series of steps into a single xunit.xml file.

    """

    def __init__(self, xunit_destination, name, package_name=None):
        self.name = name
        self.package_name = package_name
        self.destination = xunit_destination
        self.event_receiver = None


    def __enter__(self):
        self.event_receiver = EventReceiver()
        return self


    def now_seconds(self):
        return time.time()


    def step(self, step_name):
        """Start a new step. returns a context manager which allows you to
        report an error"""

        @contextmanager
        def step_context(step_name):
            if self.event_receiver.current_case is not None:
                raise Exception('cannot open a step within a step')

            self.event_receiver.begin_case(step_name, self.now_seconds(), self.name)
            try:
                yield self.event_receiver
            except:
                etype, evalue, tb = sys.exc_info()
                self.event_receiver.error('%r' % [etype, evalue, tb])
                raise
            finally:
                self.event_receiver.end_case(step_name, self.now_seconds())

        return step_context(step_name)


    def __exit__(self, *exc_info):
        results = self.event_receiver.results()
        if not results:
            already_throwing = exc_info and exc_info[0] is not None
            if not already_throwing:
                raise ValueError('your hook must at least perform one step!')

        self.destination.write_reports(
            self.name, self.name, results, package_name=self.package_name,
        )



import os

from .main import toxml

class XunitDestination(object):
    """Manages a repository of xunit files, for writing test reports"""

    def __init__(self, root_dir):
        self.root_dir = root_dir
        self.expected_xunit_files = []


    def write_reports(self, relative_path, suite_name, reports,
                      package_name=None):
        """write the collection of reports to the given path"""

        dest_path = self.reserve_file(relative_path)
        with open(dest_path, 'wb') as outf:
            outf.write(toxml(reports, suite_name, package_name=package_name))
        return dest_path


    def reserve_file(self, relative_path):
        """reserve a XML file for the slice at <relative_path>.xml

        - the relative path will be created for you
        - not writing anything to that file is an error
        """
        if os.path.isabs(relative_path):
            raise ValueError('%s must be a relative path' % relative_path)

        dest_path = os.path.join(self.root_dir, '%s.xml' % relative_path)

        if os.path.exists(dest_path):
            raise ValueError('%r must not already exist' % dest_path)

        if dest_path in self.expected_xunit_files:
            raise ValueError('%r already reserved' % dest_path)

        dest_dir = os.path.dirname(dest_path)
        if not os.path.isdir(dest_dir):
            os.makedirs(dest_dir)

        self.expected_xunit_files.append(dest_path)

        return dest_path


    def check(self):
        expected = self.expected_xunit_files
        if not all(os.path.exists(path) for path in expected):
            raise Exception(
                'result files %r reserved by hook have not been produced' % (
                    set(path for path in expected if not os.path.isfile(path))
            ))



"""convert test events to a xunit XML file"""

from xml.etree import ElementTree as et

from datetime import datetime
from socket import gethostname


class Report(object):
    """represents a test case report"""

    def __init__(self, name, start_ts=None, end_ts=None, src_location=None):
        self.name = name
        self.start_ts = start_ts
        self.end_ts = end_ts
        self.src_location = src_location
        self.failures = []
        self.errors = []

    def __repr__(self):
        return '%r' % dict(
            name=self.name,
            start_ts=self.start_ts,
            end_ts=self.end_ts,
            failed=not self.errors,
            src_location=self.src_location,
        )

    def __hash__(self):
        return repr(self).__hash__()

    def __eq__(self, another):
        return repr(self) == repr(another)


class EventReceiver(object):
    """eventfull interface to collect results from test cases and produce test reports."""

    def __init__(self):
        self.cases = []
        self.current_case = None

    def end_current_case(self, ts):
        self.current_case.end_ts = ts
        self.cases.append(self.current_case)

    def begin_case(self, test_name, ts, src_location):
        if self.current_case is not None:
            self.error(ts)
            self.end_current_case(ts)

        self.current_case = Report(test_name)
        self.current_case.start_ts = ts
        self.current_case.src_location = src_location

    def end_case(self, test_name, ts):
        if self.current_case is None or self.current_case.name != test_name:
            raise Exception(
                'cannot close case %s (current: %s)' % (test_name, self.current_case)
            )
        self.end_current_case(ts)
        self.current_case = None

    def error(self, reason):
        assert self.current_case is not None
        self.current_case.errors.append(
            'error %s' % (
                reason
            )
        )

    def failure(self, reason, src_location):
        assert self.current_case is not None
        self.current_case.failures.append(
            'test failure %s at %s' % (
                reason, src_location,
            )
        )

    def results(self):
        if self.current_case is not None:
            self.error('test finished unexpectedly')
            self.end_current_case(self.current_case.start_ts)

        return self.cases


def toxml(test_reports, suite_name,
          hostname=gethostname(), package_name="tests"):
    """convert test reports into an xml file"""

    testsuites = et.Element("testsuites")
    testsuite = et.SubElement(testsuites, "testsuite")

    test_count = len(test_reports)
    if test_count < 1:
        raise ValueError('there must be at least one test report')


    assert test_count > 0, 'expecting at least one test'

    error_count = len([r for r in test_reports if r.errors])
    failure_count = len([r for r in test_reports if r.failures])
    ts = test_reports[0].start_ts
    start_timestamp = datetime.fromtimestamp(ts).isoformat()

    total_duration = test_reports[-1].end_ts - test_reports[0].start_ts

    def quote_attribute(value):
        return value if value is not None else "(null)"


    testsuite.attrib = dict(
        id="0",
        errors=str(error_count),
        failures=str(failure_count),
        tests=str(test_count),
        hostname=quote_attribute(hostname),
        timestamp=quote_attribute(start_timestamp),
        time="%f" % total_duration,
        name=quote_attribute(suite_name),
        package=quote_attribute(package_name),
    )

    for r in test_reports:
        test_name = r.name
        test_duration = r.end_ts - r.start_ts
        class_name = r.src_location

        testcase = et.SubElement(testsuite, "testcase")
        testcase.attrib = dict(
            name=test_name,
            classname=quote_attribute(class_name),
            time="%f" % test_duration,
            )
        if r.errors or r.failures:
            if r.failures:
                failure = et.SubElement(testcase, "failure")
                failure.attrib = dict(
                    type="exception",
                    message=quote_attribute('\n'.join(['%s' % e for e in r.failures])),
                )
            else:
                error = et.SubElement(testcase, "error")
                error.attrib = dict(
                    type="exception",
                    message=quote_attribute('\n'.join(['%s' % e for e in r.errors])),
                )

    return et.tostring(testsuites, encoding="utf-8")



# -*- coding: utf-8 -*-
from distutils.core import setup

setup(
    name="xunitgen",
    version="1.0.1",
    description="Generate xUnit.xml files",
    author="Nicolas Léveillé",
    author_email="nicolas@uucidl.com",
    url="https://github.com/uucidl/uu.xunitgen",
    packages=["xunitgen"],
    license='MIT',
    long_description=open('README.rst').read(),
)
